package CalculiX;

import org.junit.Test ;
import static org.junit.Assert.*;

public class Main {

  // We have a choice between three parsers.

  // This parser is generated by JavaCUP, an LR(1) parser generator.
  // The grammar is in Parser.cup.
  static final int CUP = 0;
  // This parser is written by hand. It is an LL(1) parser.
  // Its source code is in MyLLParser.java.
  static final int HANDCRAFTED = 1;
  // This parser is generated by JavaCC, an LL(1) parser generator.
  // The grammar is in MyJavaCCParser.jj.
  static final int JAVACC = 2;

  // The first two parsers above rely on a lexer which is generated
  // by jflex from a description in CalculiXLexer.flex. The third
  // parser contains its own lexer (JavaCC produces a lexer and a
  // parser together).

  // The three parsers above are not equivalent. Only the CUP parser
  // recognizes the full CalculiX language; because I am lazy, the
  // other two parsers recognize only a subset of the language.

  // Which parser do you prefer?
  static final int preferredParser = CUP;

  // This method initializes and runs the parser.
  // It requires a Reader and an expression factory.
  // It returns an expression.

  static Expression parse (java.io.Reader reader, ExpressionFactory factory) throws Exception
  {
    switch (preferredParser) {
    case CUP:
      {
        // Set up a lexer.
        CalculiXLexer lexer = new CalculiXLexer (reader);
	// Set up the parser.
	MyParser parser = new MyParser (lexer);
	// Allow the parser to access the expression factory.
	CUP$parser$actions.factory = factory;
	// Run the parser.
	// The method parse() produces a Symbol, whose semantic value is
	// an abstract syntax tree for an expression.
	return (Expression) parser.parse().value;
      }
    case HANDCRAFTED:
      {
        // Set up a lexer.
        CalculiXLexer lexer = new CalculiXLexer (reader);
	// Set up the parser.
	MyLLParser parser = new MyLLParser (new MyScannerWrapper (lexer), factory);
	// Run the parser.
	return parser.consumeCompleteExpression();
      }
    case JAVACC:
      {
	// Set up the parser.
	MyJavaCCParser parser = new MyJavaCCParser (reader);
	// Allow the parser to access the expression factory.
	parser.setFactory(factory);
	// Run the parser.
        return parser.completeExpression();
      }
    default:
      throw new AssertionError (); // unreachable statement
    }
  }
 
  // This method runs the parser and the interpreter.
  // It requires a Reader.
  // It produces a Value.

  static Value parseAndInterpret (java.io.Reader reader) throws Exception
  {
    // Set up the factory.
    ExpressionFactory factory = new MyExpressionFactory ();
    // Set up and run the parser.
    Expression e = parse (reader, factory);
    // Interpret the expression.
    Value v = e.eval(Environment.empty);
    // Return the result.
    return v;
  }

  // The main method.
  // It reads from the standard input channel.
  // It parses an expression, evaluates it, and prints its value.

  // Note that the parser and the interpreter can throw an exception.
  // We let this exception escape.

  public static void main (String[] argv) throws Exception
  {
    // Read from the standard input channel.
    java.io.Reader reader = new java.io.InputStreamReader (System.in);
    // Parse and interpret.
    Value v = parseAndInterpret (reader);
    // Display the value.
    System.out.println(v);
  }

  // A test method.
  // It reads from a string.
  // It parses an expression, evaluates it, and compares its value with
  // the expected value (which must be an integer value).

  static void interpretAndValidate (String text, int expectedResult) throws Exception
  {
    // Read from the supplied string.
    java.io.Reader reader = new java.io.StringReader (text);
    // Parse and interpret.
    Value v = parseAndInterpret (reader);
    // Make sure that the actual result is an integer value,
    // and compare with the expected result.
    assertEquals("The result is incorrect:", expectedResult, v.asInteger());
  }

  // Individual tests.

  @Test public void test00 () throws Exception
  {
    interpretAndValidate("0", 0);
  }

  @Test public void test00bis () throws Exception
  {
    interpretAndValidate("2*3", 6);
  }

  @Test public void test01 () throws Exception
  {
    interpretAndValidate("2-3", -1);
  }

  @Test public void test02 () throws Exception
  {
    interpretAndValidate("3-2", 1);
  }

  @Test public void test03 () throws Exception
  {
    interpretAndValidate("3-2-1", 0);
  }

  @Test public void test03bis () throws Exception
  {
    interpretAndValidate("0-1-1-1-1-1-1-1-1-1-1", -10);
  }

  @Test public void test04 () throws Exception
  {
    interpretAndValidate("8/4/2", 1);
  }

  @Test public void test05 () throws Exception
  {
    interpretAndValidate("1-2+3", 2);
  }

  @Test public void test06 () throws Exception
  {
    interpretAndValidate("8/2*4", 16);
  }

  @Test public void test07 () throws Exception
  {
    interpretAndValidate("1+2*3", 7);
  }

  @Test public void test08 () throws Exception
  {
    interpretAndValidate("1+3/2", 2);
  }

  @Test public void test09 () throws Exception
  {
    interpretAndValidate("2*3-1", 5);
  }

  @Test public void test10 () throws Exception
  {
    interpretAndValidate("16/4-1", 3);
  }

  @Test public void test11 () throws Exception
  {
    interpretAndValidate("2 + 3*5 - 30/2 - 2", 0);
  }

  @Test public void test12 () throws Exception
  {
    interpretAndValidate("(5 - 2) * 2 - 24 / (2 + 2)", 0);
  }

  @Test public void test13 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("if true then 0 else 1", 0);
  }

  @Test public void test14 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("let x = 2 * 21 in x", 42);
  }

  @Test public void test15 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("let x = 2 in let y = 21 in x * y", 42);
  }

  @Test public void test16 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("let x = 2 in let x = 21 in x * x", 441);
  }

  @Test public void test17 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("let x = 2 in let y = x + 19 in x * y", 42);
  }

  @Test public void test18 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("let x = 2 in let x = x + 19 in x * x", 441);
  }

  @Test public void test19 () throws Exception
  {
    if (preferredParser != JAVACC)
      interpretAndValidate("(let x = 2 in x + x) * (let x = 3 in x + x)", 24);
  }

  @Test public void test20 () throws Exception
  {
    if (preferredParser == CUP)
      interpretAndValidate("for x = 0 to 10 sum x", 55);
  }

  @Test public void test21 () throws Exception
  {
    if (preferredParser == CUP)
      interpretAndValidate("1 + (for x = 0 to 10 sum x * x)", 386);
  }

  @Test public void test22 () throws Exception
  {
    if (preferredParser == CUP)
      interpretAndValidate("if 1+1 < 1 || 4 < 4*4 then 3 else 4", 3);
  }

  @Test public void test23 () throws Exception
  {
    if (preferredParser == CUP)
      interpretAndValidate("if !true && false then 1 else 0", 0);
  }

  @Test public void test24 () throws Exception
  {
    if (preferredParser == CUP)
      interpretAndValidate("if false && false || true then 1 else 0", 1);
  }

  // Negative tests.

  @Test(expected=Throwable.class) public void fail00 () throws Exception
  {
    // The lexer should fail because it does not accept this character.
    interpretAndValidate("#", 0);
    // The lexer produced by JavaCC throws a TokenMgrError, a subclass of
    // Error, so we must be prepared to see any Throwable object.
  }

  @Test(expected=Exception.class) public void fail01 () throws Exception
  {
    if (preferredParser == JAVACC) throw new Exception ("Skipping this test");
    // The parser should fail because an operator is missing.
    interpretAndValidate("42 x", 0);
  }

  @Test(expected=Exception.class) public void fail02 () throws Exception
  {
    // The parser should fail because an operand is missing.
    interpretAndValidate("1+42++3", 0);
  }

  @Test(expected=Exception.class) public void fail03 () throws Exception
  {
    // The parser should fail because there are too many closing
    // parentheses.  If the parser mistakenly accepts "(42)" and
    // ignores the extra parenthesis, this test will detect it.
    interpretAndValidate("(42))", 42);
  }

  @Test(expected=Exception.class) public void fail04 () throws Exception
  {
    // The parser should fail because there are too many opening parentheses.
    interpretAndValidate("((42)", 0);
  }

  @Test(expected=EvalException.class) public void fail05 () throws Exception
  {
    if (preferredParser == JAVACC) throw new EvalException ("Skipping this test");
    // The interpreter should fail because a variable is undefined.
    interpretAndValidate("x", 0);
  }

  @Test(expected=EvalException.class) public void fail06 () throws Exception
  {
    if (preferredParser == JAVACC) throw new EvalException ("Skipping this test");
    // The interpreter should fail because a variable is undefined.
    interpretAndValidate("let x = x + 1 in x", 0);
  }

  @Test(expected=EvalException.class) public void fail07 () throws Exception
  {
    if (preferredParser == JAVACC) throw new EvalException ("Skipping this test");
    // The interpreter should fail because a variable is undefined.
    interpretAndValidate("(let x = 2 in x) + x", 0);
  }

  @Test(expected=EvalException.class) public void fail08 () throws Exception
  {
    if (preferredParser == JAVACC) throw new EvalException ("Skipping this test");
    // The interpreter should fail because the operands are not integers.
    interpretAndValidate("true + true", 0);
  }

  @Test(expected=EvalException.class) public void fail09 () throws Exception
  {
    if (preferredParser == JAVACC) throw new EvalException ("Skipping this test");
    // The interpreter should fail because the condition is not a Boolean.
    interpretAndValidate("if 0 then 1 else 2", 0);
  }

}
