package CalculiX;

// This class implements a hand-written LL(1) parser.

// It relies on a scanner that implements the interface MyScanner.

// It relies on a factory that implements the interface ExpressionFactory.

import static CalculiX.sym.*;

class MyLLParser {

  // The scanner that we are using to read tokens.
  private final MyScanner scanner;

  // The factory that we are using to construct expressions.
  private final ExpressionFactory factory;

  // Constructor.

  MyLLParser (MyScanner scanner, ExpressionFactory factory)
  {
    this.scanner = scanner;
    this.factory = factory;
  }

  // Each of the methods that follow corresponds to a single non-terminal
  // symbol in our grammar.

  Expression consumeAtomicExpression () throws Exception
  {
    switch (scanner.consultNextToken()) {
    case LPAREN:
      scanner.consumeNextToken();
      Expression e = consumeExpression();
      scanner.consumeNextToken(RPAREN);
      return e;
    case IDENT:
      String x = (String) scanner.consumeNextToken();
      return factory.buildVarRead(x);
    case INTEGER:
      Integer c = (Integer) scanner.consumeNextToken();
      return factory.buildConstant(c);
    case TRUE:
      scanner.consumeNextToken();
      return factory.buildConstant(true);
    case FALSE:
      scanner.consumeNextToken();
      return factory.buildConstant(false);
    default:
      throw scanner.syntaxError ("expected an atomic expression");
    }
  }

  Operator consumeMultiplicativeOperator () throws Exception
  {
    Operator op;
    switch (scanner.consultNextToken()) {
    case STAR:
      scanner.consumeNextToken();
      return Operator.MULTIPLICATION;
    case SLASH:
      scanner.consumeNextToken();
      return Operator.DIVISION;
    default:
      throw scanner.syntaxError ("this cannot happen");
    }
  }

  Expression consumeMultiplicativeExpression () throws Exception
  {
    Expression left = consumeAtomicExpression();
    return consumeMultiplicativeExpressionTail (left);
  }

  Expression consumeMultiplicativeExpressionTail (Expression left) throws Exception
  {
    switch (scanner.consultNextToken()) {
    case STAR:
    case SLASH:
      {
	Operator op = consumeMultiplicativeOperator();
	Expression right = consumeAtomicExpression();
	Expression e = factory.buildIntegerOperation(left, op, right);
	return consumeMultiplicativeExpressionTail (e);
      }
    case PLUS:
    case MINUS:
    case RPAREN:
    case THEN:
    case ELSE:
    case IN:
    case EOF:
      return left;
    default:
      throw scanner.syntaxError ("syntax error");
    }
  }

  Operator consumeAdditiveOperator () throws Exception
  {
    Operator op;
    switch (scanner.consultNextToken()) {
    case PLUS:
      scanner.consumeNextToken();
      return Operator.ADDITION;
    case MINUS:
      scanner.consumeNextToken();
      return Operator.SUBTRACTION;
    default:
      throw scanner.syntaxError ("this cannot happen");
    }
  }

  // Here, we could follow exactly the same structure as above (see
  // consumeMultiplicativeExpression and consumeMultiplicativeExpressionTail).
  // However, because the two calls to consume*ExpressionTail are tail calls,
  // we can replace them with a while loop.

  Expression consumeAdditiveExpression () throws Exception
  {
    Expression left = consumeMultiplicativeExpression();
    while (true) {
      switch (scanner.consultNextToken()) {
      case PLUS:
      case MINUS:
        {
          Operator op = consumeAdditiveOperator();
          Expression right = consumeMultiplicativeExpression();
          left = factory.buildIntegerOperation(left, op, right);
          break;
        }
      case RPAREN:
      case THEN:
      case ELSE:
      case IN:
      case EOF:
        return left;
      default:
        throw scanner.syntaxError ("syntax error");
      }
    }
  }

  Expression consumeExpression () throws Exception
  {
    switch (scanner.consultNextToken()) {
    case LPAREN:
    case IDENT:
    case INTEGER:
    case TRUE:
    case FALSE:
      return consumeAdditiveExpression();
    case IF:
      {
	scanner.consumeNextToken();
	Expression b = consumeExpression();
	scanner.consumeNextToken(THEN);
	Expression e1 = consumeExpression();
	scanner.consumeNextToken(ELSE);
	Expression e2 = consumeExpression();
	return factory.buildIf(b, e1, e2);
      }
    case LET:
      {
	scanner.consumeNextToken();
	String x = (String) scanner.consumeNextToken(IDENT);
	scanner.consumeNextToken(EQ);
	Expression e1 = consumeExpression();
	scanner.consumeNextToken(IN);
	Expression e2 = consumeExpression();
	return factory.buildLet(x, e1, e2);
      }
    default:
      throw scanner.syntaxError ("syntax error");
    }
  }

  Expression consumeCompleteExpression () throws Exception
  {
    Expression e = consumeExpression();
    switch (scanner.consultNextToken()) {
    case EOF:
      return e;
    default:
      throw scanner.syntaxError ("expected EOF");
    }
  }

}

