package CalculiX;

// This class is used to convert an object that implements the interface
// Scanner (as produced using jflex) into one that implements MyScanner.

// MyScanner is richer and thus easier to use than Scanner.

// By convention, when the end of file is reached, the underlying scanner
// produces a symbol whose tag is EOF. Thus, we need not worry about the
// end of file.

// Note that, although the lexer is produced by jflex, the token definitions
// are produced by JavaCUP, which is still necessary. Of course, one could
// decide to manually produce these definitions.

import java_cup.runtime.Scanner;
import java_cup.runtime.Symbol;

class MyScannerWrapper implements MyScanner {

  // The wrapped scanner.
  private final Scanner scanner;

  // The next symbol of the input. By convention, this is null
  // if and only if we have not started yet.
  private Symbol nextSymbol;

  // The current position. This is used for reporting errors.
  private int position;

  // This ugly piece of code uses the Java reflection API to convert
  // a symbol tag (as defined in the class "sym") into a string.
  // This code was copied from code produced by jflex.

  private static String getTokenName (int tag) throws Exception
  {
    java.lang.reflect.Field [] classFields = sym.class.getFields();
    for (int i = 0; i < classFields.length; i++)
      if (classFields[i].getInt(null) == tag)
	return classFields[i].getName();
    throw new AssertionError (); // unreachable statement
  }

  // Constructor.

  MyScannerWrapper (Scanner scanner)
  {
    this.scanner = scanner;
    this.nextSymbol = null;
    this.position = 0;
  }

  // This method ensures that the scanner has been called at least once,
  // so the field nextSymbol holds the next input symbol.

  private void lazyStart () throws Exception
  {
    // If the lexer has not been started yet, start it now and obtain
    // the very first token.
    if (nextSymbol == null) {
      nextSymbol = scanner.next_token();
      assert (nextSymbol != null); // underlying scanner cannot return null
    }
  }    

  // This method returns the tag of the next input symbol, without
  // consuming it.

  public int consultNextToken () throws Exception
  {
    lazyStart();
    return nextSymbol.sym;
  }

  // This method returns the semantic value of the next input symbol,
  // which is consumed.

  public Object consumeNextToken () throws Exception
  {
    lazyStart();
    assert (nextSymbol.sym != sym.EOF); // EOF should not be consumed
    Object value = nextSymbol.value;
    nextSymbol = scanner.next_token();
    position++;
    assert (nextSymbol != null); // underlying scanner cannot return null
    return value;
  }

  // This method returns the semantic value of the next input symbol,
  // which is consumed. In addition, it checks that this symbol carries
  // the expected tag.

  public Object consumeNextToken (int expectedTag) throws Exception
  {
    int actualTag = consultNextToken();
    if (actualTag == expectedTag)
      return consumeNextToken();
    else
      throw syntaxError ("expected " + getTokenName(expectedTag));
  }

  // This method is used to report a syntax error. It constructs an Exception
  // object, which can then be thrown by the caller.

  public Exception syntaxError (String message)
  {
    return new Exception ("Position " + position + ": " + message);
  }

}

